export class SQLiteClient {
    constructor(db, _) {
        this.db = db;
    }
    async getCoValue(coValueId) {
        const { rows } = await this.db.execute("SELECT * FROM coValues WHERE id = ?", [coValueId]);
        if (!rows || rows.length === 0)
            return;
        const coValueRow = rows[0];
        try {
            const parsedHeader = coValueRow?.header &&
                JSON.parse(coValueRow.header);
            return {
                ...coValueRow,
                header: parsedHeader,
            };
        }
        catch (e) {
            console.warn(coValueId, "Invalid JSON in header", e, coValueRow?.header);
            return;
        }
    }
    async getCoValueSessions(coValueRowId) {
        const { rows } = await this.db.execute("SELECT * FROM sessions WHERE coValue = ?", [coValueRowId]);
        return rows;
    }
    async getNewTransactionInSession(sessionRowId, firstNewTxIdx) {
        const { rows } = await this.db.execute("SELECT * FROM transactions WHERE ses = ? AND idx >= ?", [sessionRowId, firstNewTxIdx]);
        if (!rows || rows.length === 0)
            return [];
        try {
            return rows.map((row) => ({
                ...row,
                tx: JSON.parse(row.tx),
            }));
        }
        catch (e) {
            console.warn("Invalid JSON in transaction", e);
            return [];
        }
    }
    getSignatures(sessionRowId, firstNewTxIdx) {
        const { rows } = this.db.executeSync("SELECT * FROM signatureAfter WHERE ses = ? AND idx >= ?", [sessionRowId, firstNewTxIdx]);
        return rows;
    }
    async addCoValue(msg) {
        const { insertId } = await this.db.execute("INSERT INTO coValues (id, header) VALUES (?, ?)", [msg.id, JSON.stringify(msg.header)]);
        return insertId ?? 0;
    }
    async addSessionUpdate({ sessionUpdate, }) {
        const { rows } = await this.db.execute(`INSERT INTO sessions (coValue, sessionID, lastIdx, lastSignature, bytesSinceLastSignature) 
       VALUES (?, ?, ?, ?, ?)
       ON CONFLICT(coValue, sessionID) 
       DO UPDATE SET lastIdx=excluded.lastIdx, 
                    lastSignature=excluded.lastSignature, 
                    bytesSinceLastSignature=excluded.bytesSinceLastSignature
       RETURNING rowID`, [
            sessionUpdate.coValue,
            sessionUpdate.sessionID,
            sessionUpdate.lastIdx,
            sessionUpdate.lastSignature,
            sessionUpdate.bytesSinceLastSignature,
        ]);
        return rows[0]?.rowID;
    }
    async addTransaction(sessionRowID, nextIdx, newTransaction) {
        await this.db.execute("INSERT INTO transactions (ses, idx, tx) VALUES (?, ?, ?)", [sessionRowID, nextIdx, JSON.stringify(newTransaction)]);
    }
    async addSignatureAfter({ sessionRowID, idx, signature, }) {
        await this.db.execute("INSERT INTO signatureAfter (ses, idx, signature) VALUES (?, ?, ?)", [sessionRowID, idx, signature]);
    }
    async unitOfWork(operationsCallback) {
        try {
            await this.db.transaction(async () => {
                await Promise.all(operationsCallback());
            });
        }
        catch (e) {
            console.error("Transaction failed:", e);
            throw e;
        }
    }
}
//# sourceMappingURL=client.js.map